package tools;

import inspectors.OutputInspector;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;

import junit.framework.Assert;
import asct.shared.ApplicationState;
import asct.shared.ExecutionRequestStatus;
import asct.shared.ParametricCopyHolder;
import asct.ui.ASCTController;
import asct.ui.ExecutionRequestDescriptorComponent;
import asct.ui.ExecutionRequestDescriptorComposite;
import dataTypes.ApplicationType;

public class CommonApplicationTest {

	private static final String ABSOLUTE_TEST_PATH = "/tmp/integrade_test/";

	private String basePath;

	private String applicationName;

	private String binary;

	private ASCTController asctController;

	private int numberOfTasks;

	private FinishedExecutionListener listener;

	private int applicationNumber;

	private static int applicationsCount = 0;

	private String[] inputFiles;

	private String[] outputFiles;

	private ParametricCopyHolder[] parametricCopies;

	private String applicationArguments;

	private String applicationConstraints;

	private String applicationPreferences;

	private ApplicationType applicationType;

	private boolean shouldForceDifferentMachines;

	private OutputInspector outputInspector;

	private CommonApplicationTest(String applicationName,
			int applicationNumber, ASCTController controller) {

		this.applicationName = applicationName;
		asctController = controller;
		listener = new FinishedExecutionListener();
		asctController.registerExecutionStateListener(listener);
		outputInspector = new OutputInspector();

		this.applicationNumber = applicationNumber;
	}

	public static CommonApplicationTest generate(String applicationName,
			ASCTController controller) {
		int applicationNumber = applicationsCount;
		applicationsCount++;
		return new CommonApplicationTest(applicationName, applicationNumber,
				controller);
	}

	public void setup() throws Exception {
		ExecutionRequestDescriptorComposite requestDescriptor = null;

		// Create Paths
		String configurationAbsolutePath = ABSOLUTE_TEST_PATH + applicationName
				+ ".conf";
		String applicationAbsolutePath = ABSOLUTE_TEST_PATH + applicationName;

		BufferedReader descFile = new BufferedReader(new InputStreamReader(
				new FileInputStream(configurationAbsolutePath)));
		requestDescriptor = asctController
				.readExecutionRequestDescriptorFile(descFile);

		// Load File Data
		basePath = requestDescriptor.getApplicationBasePath();
		binary = requestDescriptor.getApplicationBinaries();
		applicationType = requestDescriptor.getApplicationType();
		shouldForceDifferentMachines = requestDescriptor
				.getShouldForceDifferentMachines();
		numberOfTasks = requestDescriptor.getNumberOfTasks();

		// Upload Application to Grid
		asctController.createDirectory(basePath);
		asctController.registerApplication(basePath, applicationName);
		asctController.uploadBinary(applicationAbsolutePath, basePath,
				applicationName, binary);

		// Get Input Files
		inputFiles = getFileListFromIterator(requestDescriptor.getInputFiles());

		// Get Ouput Files
		outputFiles = getFileListFromIterator(requestDescriptor
				.getOutputFiles());

		// Load Parametric Files
		ArrayList<ParametricCopyHolder> parametricCopiesList = getParametricCopyList(requestDescriptor);
		parametricCopies = new ParametricCopyHolder[parametricCopiesList.size()];
		parametricCopiesList.toArray(parametricCopies);

		// Get Constraints and Arguments
		applicationArguments = protectFromNullReturn(requestDescriptor
				.getApplicationArguments());
		applicationConstraints = protectFromNullReturn(requestDescriptor
				.getApplicationConstraints());
		applicationPreferences = protectFromNullReturn(requestDescriptor
				.getApplicationPreferences());
	}

	private String[] getFileListFromIterator(Iterator<String> iterator) {
		ArrayList<String> fileCollection = new ArrayList<String>();
		for (Iterator<String> i = iterator; i.hasNext();) {
			fileCollection.add(i.next());
		}
		String[] fileList = new String[fileCollection.size()];
		fileCollection.toArray(fileList);
		return fileList;
	}

	private String protectFromNullReturn(String unsafeValue) {
		return unsafeValue == null ? "" : unsafeValue;
	}

	private ArrayList<ParametricCopyHolder> getParametricCopyList(
			ExecutionRequestDescriptorComposite parametricRequestDescriptor) {
		int id = 0;
		ArrayList<ParametricCopyHolder> parametricCopiesList = new ArrayList<ParametricCopyHolder>();
		for (Iterator<ExecutionRequestDescriptorComponent> i = parametricRequestDescriptor
				.getParametricApplicationCopies(); i.hasNext();) {
			ExecutionRequestDescriptorComponent appCopy = i.next();

			String[] inputArray = getFileListFromIterator(appCopy
					.getInputFiles());
			String[] outputArray = getFileListFromIterator(appCopy
					.getOutputFiles());
			ParametricCopyHolder pc = new ParametricCopyHolder(id++, appCopy
					.getApplicationArguments(), inputArray, outputArray);
			parametricCopiesList.add(pc);
		}
		return parametricCopiesList;
	}

	public void test() throws Exception {
		run(1000);

		HashMap<String, String> hash = outputInspector
				.inspectOutput(asctController.getOutputDirectory() + "/"
						+ String.valueOf(applicationNumber) + "/0/stdout");
		HashMap<String, String> expectedHash = outputInspector
				.inspectOutput(ABSOLUTE_TEST_PATH + applicationName
						+ ".expected");

		Assert.assertEquals(expectedHash, hash);
	}

	public void tearDown() {
		String message = "";
		boolean failed = false;
		try {
			asctController.deleteBinary(basePath, applicationName, binary);
		} catch (Exception e) {
			message += " delete";
			failed = true;
		}

		try {
			asctController.unregisterApplication(basePath, applicationName);

		} catch (Exception e) {
			message += " unregister";
			failed = true;
		}
		try {
			asctController.removeDirectory(basePath);

		} catch (Exception e) {
			message += " remove";
			failed = true;
		}

		if (failed == true)
			Assert.fail("Cant perform cleaning actions: " + message);

	}

	public ASCTController getController() {
		return asctController;
	}

	public String getName() {
		return applicationName;
	}

	public Integer getId() {
		return applicationNumber;
	}

	public ExecutionRequestStatus run(long timeout) throws Exception {
		ExecutionRequestStatus status = null;
		status = asctController.executeApplication(basePath, applicationName,
				applicationArguments, applicationConstraints,
				applicationPreferences, applicationType, binary, inputFiles,
				outputFiles, parametricCopies, numberOfTasks,
				shouldForceDifferentMachines);

		int tries = 0;
		while (!listener.hasFinished()) {
			try {
				Thread.sleep(timeout);
				if (tries++ > 5) {
					Assert.fail("Timeout reached: " + tries + "s");
				}
			} catch (InterruptedException e) {
				Assert.fail("ThreadJVM can't sleep");
				e.printStackTrace();
			}
		}

		// FIXME If status is null, it should be returned as a unknown status
		return status;

	}
}
